'''
Constants for the core module.
'''

root = 'https://neko-sama.fr'

headers = {
    'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8',
    'Accept-Encoding': 'gzip, deflate, br',
    'Accept-Language': 'fr,fr-FR;q=0.8,en-US;q=0.5,en;q=0.3',
    'Connection': 'keep-alive',
    'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:2.0.1) Gecko/20100101 Firefox/4.0.1'
}

class re:
    '''
    Useful parsing regexes.
    TODO refactor names
    '''
    
    ep_list    = r'var episodes = ([^;]*)'                              # Get eps list
    providers  = r"video\[\d\] = '.*';"                                 # Get ep providers
    script     = r'https://[a-z.]*\/[a-z\/-]*\?.*\" '                   # Get backend script
    m3u8       = r'\"([a-zA-Z-0-9=]{25,})\"'                            # Get M3U8 info
    m3u8_urls  = r'https.*\"'                                           # Get M3U8 urls
    qualities  = r'NAME=\"\d*\"\nhttps://[^#]*'                         # Get quality urls
    fragments  = r'https://.*'                                          # Get fragment urls
    ep_id      = r'https://.*/episode/(\d*)'                            # Get episode id
    ep_index   = r'https:\/\/.*\/episode\/.*-(\d*)_.*'                  # Get episode index
    ep_name    = r'https:\/\/.*\/episode\/\d*-(.*)_'                    # Get episode name
    valid_ep   = r'https://neko-sama\.fr/anime/episode/\d*[a-z-\d_]*'   # Validate an episode
    frag_index = r'\/(\d*)\.ts'                                         # Get fragment index
    ani_tags   = r'&quot;(.*?)&quot;'                                   # Get anime search tags
    an_from_ep = r'href=\"(/anime/info/.*)\" class=\"cover\"'           # Get ani url from ep page
    ep_bg      = r'\"head\" style=\"background-image: url\((https://.*)\)' # Get ep bg image url
    
    glob_name  = r'https://.*/anime/(info|episode)/(\d*)-([a-z-\d]*)_(vostfr|vf)' # Get url data
    ani_props  = r'<meta property=\"og:(?P<name>.*)\" content=\"([.\s\S]*?)\" />' # Get meta props
   

class provider:
    '''
    Representations of providers.
    Possible values: BEST, FUSE, PSTREAM.
    '''
    
    BEST = 'best'
    FUSE = 'fusenet'
    PSTREAM = 'pstream'

class quality:
    '''
    Quality presets. Can also
    be an int instead.
    '''
    
    BEST = 'best'
    HALF = 'half'
    WORST = 'worst'

class genre:
    thriller = 'thriller'
    fantasy = 'fantasy'
    hentai = 'hentai'
    mystery = 'mystery'
    ecchi = 'ecchi'
    mahou_shoujo = 'mahou shoujo'
    shoujo = 'shoujo'
    adventure = 'adventure'
    romance = 'romance'
    shounen = 'shounen'
    slice_of_life = 'slice of life'
    horror = 'horror'
    sports = 'sports'
    isekai = 'isekai'
    magic = 'magic'
    mafia = 'mafia'
    military = 'military'
    yuri = 'yuri'
    mecha = 'mecha'
    psychological = 'psychological'
    drama = 'drama'
    cyberpunk = 'cyberpunk'
    music = 'music'
    action = 'action'
    sci_fi = 'sci-fi'
    battle_royale = 'battle royale'
    supernatural = 'supernatural'
    comedy = 'c0m1d'

class type:
    tv = 'tv'
    ova = 'ova'
    movie = 'm0v1e'
    special = 'special'
    
    unknown = ''

# EOF


'''
API wrapper for the neko-sama website.

https://github.com/Egsagon/neko-sama-api
'''


from nekosama.core import (
    Episode,
    Anime,
    Client
)

from nekosama.consts import (
    provider,
    quality,
    genre,
    type
)


'''
List of backends used by the 
core module to download animes.
'''

import os
import re
import time
import copy

import shutil
import requests
import threading

from typing import Callable
from nekosama import consts
from nekosama import utils

# Get the FFMPEG path
FFMPEG = shutil.which('ffmpeg')

BACKENDS = [
    'ffmpeg',
    'thread',
    'thread_ffmpeg',
    'safe'
]

def reach(method: str) -> Callable:
    '''
    Reach a backend function.
    '''
    
    if not method in BACKENDS:
        raise Exception('Invalid backend', method)
    
    if FFMPEG is None and 'ffmpeg' in method:
        print('[ BK ] FFMPEG is not installed, falling back to safe')
        method = 'safe'
    
    return eval('bk_' + method)


def bk_ffmpeg(raw: str,
              path: str,
              callback: Callable = None,
              quiet: bool = False,
              **kwargs) -> None:
    '''
    Download using the ffmpeg download feature.
    '''

    # Get the total count
    lenght = len(re.findall(consts.re.fragments, raw))

    # Write the m3u8 file for ffmpeg
    with open('temp.m3u', 'w') as file: file.write(raw)
    
    # Create the FFMPEG command
    command = [
        FFMPEG,
        '-protocol_whitelist', 'file,http,https,tcp,tls,crypto',
        '-i', 'temp.m3u',
        '-acodec', 'copy',
        '-vcodec', 'copy',
        path, '-y'
    ]
    
    def log(line: str) -> None:
        # Decide whether to execute the callback or not
        # for each line.
        
        key = re.findall(consts.re.frag_index, line)
        
        if len(key) and callback is not None:
            callback('downloading', int(key[0]) + 1, lenght)

        if not quiet:
            print(line, end = '')
    
    # Start FFMPEG
    assert utils.popen(command, log) == 0
    
    # Delete the m3u8 file
    os.remove('temp.m3u')

def dl_frag(req: requests.PreparedRequest,
            timeout: float,
            session: requests.Session,
            data: dict = None,
            todo: list = None) -> bytes | None:
    '''
    Download a single fragment.
    '''
    
    raw = session.send(req).content
    
    while b'<html>' in raw:
        
        # If request fails, put it back in todo
        if todo is not None:
            return todo.append(req)
        
        time.sleep(timeout)
        raw = session.send(req).content
    
    else:
        if data is None: return raw
        key = int(re.findall(consts.re.frag_index, req.url)[0])
        data[key] = raw

def bk_base_thread(raw: str,
                   timeout: float = .05,
                   callback: Callable = None,
                   quiet: bool = False,
                   **kwargs) -> dict[int, bytes]:
    '''
    Download each fragment using threads.
    Theorically faster.
    Instead of writing it to a path, returns it.
    '''
    
    chunks = re.findall(consts.re.fragments, raw)
    reqs = [requests.Request('GET', chunk, consts.headers).prepare() for chunk in chunks]
    session = requests.session()
    
    data = {}
    todo = copy.deepcopy(reqs)
    
    while len(todo):
        current = todo.pop(0)
        
        t = threading.Thread(
            target = dl_frag,
            args = [current, timeout, session, data, todo]
        )
        
        t.start()
        
        time.sleep(timeout)
        
        # Debug
        if callback is not None: callback('downloading', len(data), len(chunks))
        
        if not quiet:
            print(f'[ BK ] Downloading chunk ({len(todo)} left)')
    
    # Check for missing chunks
    for i in range(len(chunks)):
        if not i in data.keys():
            
            # Debug
            if not quiet: print('Downloading missing', i)
            
            if callback is not None:
                callback('downloading missing', i, len(chunks))
            
            # Get chunk request
            mreq = [r for r in reqs if r.url.endswith(f'{i}.ts')][0]
            data[i] = dl_frag(mreq, timeout, session)
    
    return data

def bk_thread(raw: str,
              path: str,
              timeout: float = .05,
              callback: Callable = None,
              quiet: bool = False,
              **kwargs) -> None:
    '''
    Download each fragment using threads.
    Theorically faster.
    '''
    
    # Fetch the chunks
    chunks = bk_base_thread(raw = raw,
                            timeout = timeout,
                            callback = callback,
                            quiet = quiet)
    
    # Write them to the file
    with open(path, 'wb') as output:
        for key in sorted(chunks.keys()):
            
            if not quiet:
                print(f'\r[ BK ] Downloading: {key}/{len(chunks)}', end = '')
            
            if callback is not None:
                callback('writing', key, len(chunks))
            
            output.write(chunks[key])
        
        if not quiet: print()
    
    return
    
def bk_thread_ffmpeg(raw: str,
                     path: str,
                     timeout: float = .05,
                     callback: Callable = None,
                     quiet: bool = False,
                     **kwargs) -> None:
    '''
    Same as thread, but uses ffmpeg to concatenate.
    '''
    
    # Fetch the chunks
    chunks = bk_base_thread(raw = raw,
                            timeout = timeout,
                            callback = callback,
                            quiet = quiet)
    
    temp = 'temp/'
    track = []
    
    # Create temp folder if needed
    if not os.path.exists(temp): os.mkdir(temp)
    
    for index, chunk in chunks.items():
        
        frag_path = temp + str(index) + '.ts'
        
        with open(frag_path, 'wb') as frag:
            frag.write(chunk)
        
        if not quiet:
            print(f'\r[ BK ] Writing {index}/{len(chunks)}', end = '')
        
        if callback is not None:
            callback('writing', index, len(chunks))
        
        track += [f'file {index}.ts']
    
    with open(temp + 'track', 'w') as file:
        file.write('\n'.join(track))
    
    if not quiet: print('[ BK ] Concatenating')
    
    # Concatenate ts files with ffmpeg
    
    command = [
        FFMPEG, '-f', 'concat',
        '-safe', '0',
        '-i', temp + 'track',
        '-c', 'copy', path, '-y'
    ]
    
    def log(line: str) -> None:
        # Decide to execute the callback
        # for a line or not.
        
        if line.startswith('\nframe='):
            cur = line[8:].split()[0]
            
            if callback is not None:
                callback('concatenating', int(cur), None)
        
        if not quiet:
            print(line, end = '')
    
    assert utils.popen(command, log) == 0
    
    if not quiet: print('Deleting temp files')
    
    # Delete the temp files
    for file in os.listdir(temp):
        os.remove(temp + file)

def bk_safe(raw: str,
            path: str,
            timeout: float = 0,
            callback: Callable = None,
            quiet: bool = False,
            **kwargs) -> None:
    '''
    Download one segment at a time, and concatenate their
    bytes before writing to a file.
    '''
    
    fragments = re.findall(consts.re.fragments, raw)
    session = requests.Session()
    
    with open(path, 'wb') as output:
        for i, url in enumerate(fragments):
            
            if callback: callback('downloading', i, len(fragments))
            
            if not quiet:
                print('Downloading', i)
            
            output.write(
                session.get(url, headers = consts.headers).content
            )
            
            time.sleep(timeout)
    
    print()

# EOF



'''
Utilities for the core module.
'''

import re
import os
import subprocess
from nekosama import consts

from typing import Callable
from string import ascii_letters, digits


def get_closest_value(iter: list[int], value: int):
    '''
    Pick the closest value in a list.
    From www.entechin.com/find-nearest-value-list-python/
    '''
    
    difference = lambda input_list: abs(input_list - value)
    return min(iter, key = difference)

def select_provider(prov_list: list[str], prov: str) -> str:
    '''
    Select the best provider among all of an episode's providers.
    
    Arguments
        prov_list: list of available providers.
        prov: the desired provider.
    
    Returns the url of the provider.
    '''
    
    assert len(prov_list), 'Could not fetch any provider'
    
    # Use the first of the list, usually the best one
    if prov == consts.provider.BEST:
        return prov_list[0]
    
    # Try to find the provider in the list
    for url in prov_list:
        if prov in url: return url
    
    # Fallback to best if the provider was not found
    return prov_list[0]

def select_quality(qual_list: dict[int, str], qual: str | int) -> str:
    '''
    Select the best quality among all of an episode's provider
    qualities (usually 1080, 720 or 480).
    
    Arguments
        qual_list: list of parsed qualities.
        qual: the desired quality.
        
    Returns the url of the m3u8 file.
    '''
    
    keys = list(qual_list.keys())
    
    match qual:
        
        case consts.quality.BEST:   key = max(keys)
        case consts.quality.WORST:  key = min(keys)
        case consts.quality.HALF:   key = keys[len(keys) // 2]
        
        case ukn:
            if isinstance(qual, int):
                key = get_closest_value(keys, ukn)
            
            else: raise TypeError('Invalid quality', qual)
    
    return qual_list[key]

def parse_qualities(raw: str) -> dict[int, str]:
    '''
    Parse a raw m3u8 list of qualities to a dict.
    format: {quality: m3u8_url}
    
    Arguments
        raw: a string containing the content of the m3u8 file.
    
    Returns a dict linking the qualities to their urls.
    '''
    
    quals = [q.split() for q in re.findall(consts.re.qualities, raw)]
    return {int(''.join(re.findall(r'\d*', q))): u for q, u in quals}

def ghost_format(string: str, **kwargs) -> str:
    '''
    Equivalent of str.format with ghost arguments.
    '''
    
    res = string
    
    for key, value in kwargs.items():
        if '{' + key + '}' in string:            
            res = res.replace('{' + key + '}', str(value))
    
    return res

def get_title(url: str) -> str:
    '''
    Get the title of an anime or episode given its url.
    
    Arguments
        url: the url of the anime or episode
    
    NOTE - Unused.
    '''
    
    raw = re.findall(consts.re.ep_name, url)[0]
    return ' '.join(raw.split('-')).title()

def parse_url(url: str) -> dict[str]:
    '''
    Parse informations from an anime or episode url.
    
    Arguments
        url: the anime or episode url.
    
    Returns a dict were:
        type: the type of the url (anime or episode).
        id: the id of the anime.
        name: the name of the anime.
        lang: VOSTFR or VF.
    '''
    
    infos = re.findall(consts.re.glob_name, url)[0]
    res = {k: v for k, v in zip(['type', 'id', 'name', 'lang'], infos)}
    
    if res['type'] == 'info': res['type'] = 'anime'
    return res

def popen(args: list[str], callback: Callable[[str], None]) -> int:
    '''
    Execute a shell command. Each time a line is outputed,
    call the passed callback.
    Returns the returned code.
    '''
    
    proc = subprocess.Popen(
        args,
        stdout = subprocess.PIPE,
        stderr = subprocess.STDOUT,
        text = True
    )
    
    # TODO Refactor
    
    current_line = ''
    
    while True:
        out = proc.stdout.read(1)
        
        if out == '' and proc.poll() != None: break
        
        if out != '':
            if (len(current_line) and \
                current_line[-1] == '\\' and \
                out == 'n' ) or out == '\n':
                
                callback(current_line)
                current_line = ''
            
            current_line += out
            
    return proc.poll()

def fbuild(filter: Callable | str | int | float) -> Callable[[str], bool]:
    '''
    Build a filter.
    '''
    
    # Bypass if already a function
    if callable(filter): return filter
    
    # Always pass if unspecified
    if filter is None: return lambda v: True
    
    # Absolute numbers filter
    if isinstance(filter, int | float):
        return lambda v: float(v) == filter
    
    # Relative evaluations filter
    if isinstance(filter, str):
        
        # Absolute position filter
        if '{}' in filter:
            return lambda v: eval(filter.format(v))
        
        # Left handed filter
        if filter[0] in '<=>':
            return lambda v: eval(v + filter)
        
        # Right handed filter
        if filter[-1] in '<=>':
            return lambda v: eval(filter + v)
    
    raise TypeError('Invalid filter:', filter)

# EOF


'''
Core module for the API.
'''

import re
import json
import base64
import requests

from nekosama import utils
from nekosama import consts
from nekosama import download

from time import sleep
from typing import Callable


class Episode:
    def __init__(self,
                 url: str,
                 data: dict = None,
                 anime: 'Anime' = None,
                 session: requests.Session = None) -> None:
        '''
        Represents an Anime episode.
        
        
        Arguments
            url: the Episode url.
        '''
        
        self.url = url
        self.anime = anime
        self.session = session or requests.Session()
        
        self.cache = {}
        self.providers = []
        
        self.cached_data = data
        
        self.info = utils.parse_url(self.url)
        
        # Error protection
        if self.info['type'] != 'episode':
            raise Exception('Invalid URL', url)
        
        self.id = self.info['id']
        self.name = self.info['name']
        self.lang = self.info['lang'].upper()
    
    def __str__(self) -> str:
        return f'<Episode {self.title} (ep.{self.index})>'
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def get(self,
            url: str,
            headers: dict = None,
            cache: bool = True) -> requests.Response:
        '''
        Get a page from the cache or fetch it.
        
        Arguments
            url: the url to request
            headers: the headers of the request
            cache: whether to use cache for fetching.
        '''
        
        if cache and url in self.cache:
            return self.cache.get(url)
        
        req = self.session.get(url, headers = headers)
        
        # Error protection
        if not req.ok:
            raise ConnectionError(req.status_code, req.content)
        
        self.cache[url] = req
        return req
    
    def clear_cache(self, all: bool = True) -> None:
        '''
        Clear the instance cache.
        
        Arguments
            all: whether to clear all caches.
                 if False, the data cache will be kept.
        '''
        
        # Clear request and urls caches
        self.cache.clear()
        self.providers.clear()
        
        # Clear the data cache
        if not all:
            self.cached_data.clear()
    
    @property
    def data(self) -> dict[str]:
        '''
        Get the available data for this episode.
        '''
        
        if self.cached_data is not None:
            return self.cached_data
        
        anime_url = None
        
        if (ani := self.anime) is not None:
            # Get the anime url from superior
            anime_url = ani.url
        
        else:
            # Find the anime url in the page
            raw = self.get(self.url).text
            url = re.findall(consts.re.an_from_ep, raw)[0]
            anime_url = consts.root + url
        
        src = self.anime.get(anime_url).text
        data = json.loads(re.findall(consts.re.ep_list, src)[0])
        
        # Search the episode
        for ep in data:
            if ep['url'] in self.url:
                self.cached_data = ep
                return ep
        
        raise Exception('Failed to get episode data')
    
    @property
    def time(self) -> int:
        '''
        The episode duration (in minutes).
        '''
        
        if self.data is None: self.get_data()
        return int(self.data['time'].split(' ')[0])
    
    @property
    def title(self) -> str:
        '''
        The episode title. For path naming,
        use Episode.name.
        '''
        
        if self.data is None: self.get_data()
        return self.data['title']
    
    @property
    def index(self) -> int:
        '''
        The episode index in the anime.
        '''
        
        # NOTE - We could use the get_date method
        # but that can also be done with a regex.
        
        # if self.data is None: self.get_data()
        # return self.data['num']
        
        return int(re.findall(consts.re.ep_index, self.url)[0])
    
    @property
    def image(self) -> str:
        '''
        The image url.
        '''
         
        if self.data is None: self.get_data()
        return self.data['url_image']
    
    def download_image(self, path: str) -> None:
        '''
        Download the episode image.
        
        Arguments
            path: the path to download to.
        '''
        
        raw = self.get(self.image).content
        
        with open(path, 'wb') as output:
            output.write(raw)
    
    def get_providers(self) -> list[str]:
        '''
        Get all the available providers for this episode.
        '''
        
        raw = self.get(self.url).text
        
        self.providers = [url.split('= ')[1][1:-2]
                          for url in re.findall(consts.re.providers, raw)]
        
        return self.providers

    def get_fragments(self,
                      provider: str = consts.provider.BEST,
                      quality: str | int = consts.quality.BEST) -> str:
        '''
        Get the raw m3u8 qualities list for this episode.
        '''
        
        # Get provider url (link to the player)
        if not len(self.providers): self.get_providers()
        provider_url = utils.select_provider(self.providers, provider)
        print('[FRAG] Fetched provider', provider)
        
        # Get the provider link to its script
        src = self.get(provider_url , headers = consts.headers).text
        res = re.findall(consts.re.script, src)[0][:-2]
        
        # Get the m3u file url
        src = self.get(res, headers = consts.headers).text
        res = re.findall(consts.re.m3u8, src)[0]
        raw = base64.b64decode(res).decode()
        m3u = re.findall(consts.re.m3u8_urls, raw)[0].replace('\\', '')[:-1]
        
        print('[FRAG] Fetched m3u file')
        
        # Get the m3u file data
        src = self.get(m3u, headers = consts.headers).text
        url = utils.select_quality(utils.parse_qualities(src), quality)
        
        # Fetch the fragment list
        return self.get(url, headers = consts.headers).text
    
    def download(self,
                 path: str,
                 provider: str = consts.provider.BEST,
                 quality: str = consts.quality.BEST,
                 method: str = 'ffmpeg',
                 **kwargs) -> None:
        '''
        Download the episode at a specific path.
        
        Arguments
            path: the path to download to
            provider: the provider to use (constant)
            quality: the quality to use (constant)
            method: the backend to use
            kwargs: backend arguments
        
        Quality type:
            if <int>: get the nearest quality
            if <str>: same as constants
        
        Path formating options:
            {name}: name of the episode
            {index}: index of the episode
            {id}: neko sama id of the episode
        '''
        
        path = utils.ghost_format(
            path,
            name = self.name,
            index = self.index,
            id = self.id
        )
        
        # Get the fragments list
        raw = self.get_fragments(provider, quality)
        
        # Start the download
        backend = download.reach(method)
        # print('Using backend', backend)
        
        try:
            print(f'[ EP ] Downloading episode {self.index}')
            backend(raw = raw, path = path, **kwargs)
        
        except Exception as err:
            
            # TODO
            raise err

    def to_dict(self, lazy: bool = False) -> dict:
        '''
        Return a serialized version of the object.
        TODO - lazy
        '''
        
        keys = ['name', 'title', 'image', 'time', 'id', 'lang', 'url']
        
        return {key: getattr(self, key) for key in keys}

class Anime:
    def __init__(self,
                 url: str,
                 session: requests.Session = None) -> None:
        '''
        Represents an anime containing one or multiple episodes.
        
        Arguments
            url: the anime page url
        '''
        
        self.url = url
        self.info = utils.parse_url(self.url)
        
        # Error protection
        if self.info['type'] != 'anime':
            raise Exception('Invalid url', url)
        
        self.name = self.info['name']
        self.id = self.info['id']
        self.lang = self.info['lang']
        
        self.session = session or requests.Session()
        self.cache = {}
        
        self.cached_episodes = []
    
    def __str__(self) -> str:
        # return f'<Anime `{self.title}`>'
        return f'<Anime `{self.name}`>'
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def get(self,
            url: str,
            headers: dict = None,
            cache: bool = True) -> requests.Response:
        '''
        Get a page from the cache or fetch it.
        
        Arguments
            url: the url to request
            headers: the headers of the request
            cache: whether to use cache for fetching.
        '''
        
        if cache and url in self.cache:
            return self.cache.get(url)
        
        req = self.session.get(url, headers = headers)
        
        # Error protection
        if not req.ok: raise ConnectionError(req.status_code, req.content)
        
        self.cache[url] = req
        return req
    
    def clear_cache(self, all: bool = True) -> None:
        '''
        Clear the instance cache.
        
        Arguments
            all: whether to clear all caches.
                 if False, the episodes cache will be kept.
        '''
        
        self.cache.clear()
        if not all: self.cached_episodes.clear()
    
    @property
    def data(self) -> dict[str]:
        '''
        Get anime useful data.
        '''
        
        raw = self.get(self.url).text
        return {k: v for k, v in re.findall(consts.re.ani_props, raw)}
    
    @property
    def title(self) -> str:
        '''
        The anime title.
        '''
        
        return self.data['title'].replace(' | Neko-sama', '')
    
    @property
    def description(self) -> str:
        '''
        The raw anime description.
        '''
        
        return self.data['description']
    
    @property
    def tags(self) -> list[str]:
        '''
        Identification tabs for searching.
        '''
        
        raw = self.get(self.url).text
        tags = set(re.findall(consts.re.ani_tags, raw))
        tags.remove('genres')
        
        return list(tags)
    
    @property
    def image(self) -> str:
        '''
        The image url.
        '''
         
        if self.data is None: self.get_data()
        return self.data['image']
    
    def download_image(self, path: str) -> None:
        '''
        Download the anime image.
        
        Arguments
            path: the path to download to.
        '''
        
        raw = self.get(self.image).content
        
        with open(path, 'wb') as output:
            output.write(raw)
    
    def download_background(self, path: str) -> None:
        '''
        Download the background image of the anime.
        
        Arguments
            path: the path to download to.
        '''
        
        src = self.get(self.url).text
        url = re.findall(consts.re.ep_bg, src)[0]
        raw = self.get(url).content
        
        with open(path, 'wb') as output:
            output.write(raw)
    
    def get_episodes(self) -> list[Episode]:
        '''
        Get all the anime episodes.
        '''
        
        src = self.get(self.url).text
        data = json.loads(re.findall(consts.re.ep_list, src)[0])
        
        eps = []
        for ep in data:
            eps += [Episode(url = consts.root + ep['url'],
                            data = ep,
                            anime = self,
                            session = self.session)]
        
        return eps

    @property
    def episodes(self) -> list[Episode]:
        '''
        Same as get_episodes(),
        but uses cache instead.
        '''
        
        if not len(self.cached_episodes):
            self.cached_episodes = self.get_episodes()
        
        return self.cached_episodes

    def download(self,
                 directory: str,
                 name_format: str = '{name}.mp4',
                 provider: str = consts.provider.BEST,
                 quality: str = consts.quality.BEST,
                 method: str = 'ffmpeg',
                 timeout = 5,
                 start: int = 0,
                 end: int = 0,
                 **kwargs) -> list[str]:
        '''
        Download all the episodes to a directory
        and return all the paths.
        
        Arguments
            directory: the folder to download to.
            name_format: see Episode.download.
            provider, quality, method: backend args.
            timeout: pause between each download.
        
        Returns a list of pathes.
        '''
        
        pathes = []
        
        print(f'[ AN ] Downloading anime {self.title}')
        
        for episode in self.episodes[start:end]:
            
            path = episode.download(
                path = directory + name_format,
                provider = provider,
                quality = quality,
                method = method,
                **kwargs
            )
            
            pathes += [path]
            sleep(timeout)
        
        return pathes

    def to_dict(self, lazy: bool = False) -> dict:
        '''
        Return a serialised version of the object.
        TODO - lazy
        '''
        
        return {
            'url': self.url,
            'name': self.name,
            'id': self.id,
            'title': self.title,
            'lang': self.lang,
            'episodes': len(self.episodes),
            'image': self.image,
            'tags': self.tags,
            'description': self.description,
        }
            

class Client:
    def __init__(self) -> None:
        '''
        Wrapper for both Anime and Episode classes.
        '''
        
        self.session = requests.Session()
        self.cache = {}
    
    def get(self,
            url: str,
            headers: dict = None,
            cache: bool = True) -> requests.Response:
        '''
        Get a page from the cache or fetch it.
        
        Arguments
            url: the url to request
            headers: the headers of the request
            cache: whether to use cache for fetching.
        '''
        
        if cache and url in self.cache:
            return self.cache.get(url)
        
        req = self.session.get(url, headers = headers)
        
        # Error protection
        if not req.ok: raise ConnectionError(req.status_code, req.content)
        
        self.cache[url] = req
        return req
    
    def clear_cache(self) -> None:
        '''
        Clear the instance cache.
        In particular, erase the list of episodes
        to search in.
        '''
        
        self.cache.clear()
    
    def get_anime(self, url: str) -> Anime:
        '''
        Fetch an anime from the website.
        
        Arguments
            url: the url of the anime.
        '''
        
        return Anime(url, self.session)

    def search(self,
               name: str | re.Pattern = '', # TODO pattern
               lang: str | tuple = None,
               type: list[str] = [],
               genres: list[str] = [],
               score: str = None,
               date: str = None,
               limit: int = None,
               iter_limit: int = None,
               filter: Callable[[dict], bool] = None) -> list[Anime]:
        '''
        Search for animes that match a query.
        
        Arguments
            name: the name of the anime or a part of it.
            lang: VOSTFR or VF or a tuple of both.
            type: a list of anime tags ('m0v1e', etc.)
            genres: a list of genres (shounen, action, etc.).
            score: score filter (e.g. '>4').
            date: release date filter (e.g. '>2010' or '2000<{}<2010').
            limit: stop after n animes.
            filter: callable that handle further filtering.
        
        Returns a list of Anime objects.
        '''
        
        # Lang is needed first to retrieve the files
        if lang is None: lang = ('VOSTFR', 'VF')
        if isinstance(lang, str): lang = (lang, )
        
        urls = [consts.root + f'/animes-search-{d.lower()}.json'
                for d in lang]
        
        # Fetch the urls and assemble the file
        file = []
        for url in urls:
            file += json.loads(self.session.get(url).text)
        
        # Build filters
        fdate = utils.fbuild(date)
        fscore = utils.fbuild(score)
        
        matches = []
        for index, anime in enumerate(file):
            
            # Custom filter override
            if filter and filter(anime):
                matches += [anime]
                continue
            
            # Iteration limit
            if iter_limit and index >= iter_limit: break
            
            # Result limit
            if limit and len(matches) >= limit: break
            
            # Type filter
            t = anime['type'].lower()
            if len(type) and t and t in type: continue
            
            # Genres filter
            if len(genres):
                for genre in anime['genres']:
                    if not genre in genres: continue
            
            # Score filter
            if not fscore(anime['score']): continue
            
            # Date filter
            if not fdate(anime['start_date_year']): continue
            
            # Check tags
            keys = ['title', 'title_english', 'title_romanji', 'title_french']
            
            for key in keys:
                if name in (anime[key] or '').lower():
                    matches += [anime]
                    break
        
        return [Anime(consts.root + data['url'], self.session) for data in matches]

# EOF